{
Decompresses Y. Collet's LZ4 compressed stream data in 16-bit real mode.
Optimized for 8088/8086 CPUs.
Code by Trixter/Hornet (trixter@oldskool.org) on 20130105

This specific code is probably not what you want; it is a test harness
for the code in LZ4_8088.ASM.  See that file, as well as LZ4_8088.TXT
for additional background and details.
}

program lz4decomp;

uses
  ztimer;

{
Speed optimization history (decompression times in microseconds):
before segment fixups: shuttle 108976 text 48742 broken code, invalid output
 after segment fixups: shuttle 112494 text 50940                 -
 after match copy opt: shuttle 110971 text 49890                 +
 after misc opt:       shuttle 109707 text 49056                 +
 after rep stosb opt:  shuttle 104877 text 51435                 +
 after rep stosw opt:  shuttle 104918 text 51412 robotron 365292 -+*
 after match+RLE opt:  shuttle  94274 text 49641 robotron 345426 +++
 after token unpack:   shuttle  93418 text 49140 robotron 342696 +++
 after accum opt:      shuttle  91992 text 48213 robotron 336635 +++
 after dx regswap opt: shuttle  90461 text 47218 robotron 330449 +++
 after repmovsb only:  shuttle  96231 text 46472 robotron 333068 -+- aborted
 after 1-byteRLE only: shuttle  96201 text 46472 robotron 333270 -+- aborted
 after cmp cl, -> al,: shuttle  90091 text 46894 robotron 327713 +++
 after likely(ll<15):  shuttle  89378 text 46487 robotron 323677 +++
 after ll=0 removechk: shuttle  90880 text 47957 robotron 323375 --+ aborted
 after likely(ml<15):  shuttle  89205 text 45388 robotron 317959 +++
 after mov r,ax->xchg: shuttle  88462 text 44963 robotron 315099 +++
 after es:movsw:       shuttle  90408 text 45295 robotron 321030 --- aborted
 after mcopy shortcir: shuttle  89710 text 45597 robotron 319660 --- aborted
 after rep es: movsb:  shuttle  88907 text 45076 robotron 316138 --- aborted
 after main lp unroll: shuttle  86153 text 43502 robotron 307923 +++
Peter Ferrie is credited with the following suggestions/speedups:
 remove unnecess. xor: shuttle  85781 text 43487 robotron 307660 +++
 xor ax,ax->xchg ax,r: shuttle  85037 text 43035 robotron 304574 +++
Terje Mathisen is credited with the following suggestions/speedups:
 RLE overshoot->adjus: shuttle  85022 text 43035 robotron 304571 +0+
}

function lz4_decompress(inb,outb:pointer):word; assembler;
{
Decompresses an LZ4 stream file with a compressed chunk 64K or less in size.
Input:
  DS:SI Location of source data.  DWORD magic header and DWORD chunk size
        must be intact; it is best to load the entire LZ4 file into this
        location before calling this code.

Output:
  ES:DI Decompressed data.  If using an entire 64K segment, decompression
        is "safe" because overruns will wrap around the segment.
  AX    Size of decompressed data.

Trashes AX, BX, CX, DX, SI, DI
        ...so preserve what you need before calling this code.
}
asm
        jmp     @decompinit
@SHR4table:
        DB $00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00
        DB $01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01,$01
        DB $02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02,$02
        DB $03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03,$03
        DB $04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04,$04
        DB $05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05,$05
        DB $06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06,$06
        DB $07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07,$07
        DB $08,$08,$08,$08,$08,$08,$08,$08,$08,$08,$08,$08,$08,$08,$08,$08
        DB $09,$09,$09,$09,$09,$09,$09,$09,$09,$09,$09,$09,$09,$09,$09,$09
        DB $0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A,$0A
        DB $0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B,$0B
        DB $0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C,$0C
        DB $0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D,$0D
        DB $0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E,$0E
        DB $0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F,$0F

@decompinit:
        push    ds              {preserve compiler assumptions}
        push    bp              {preserve compiler assumptions}
        les     di,outb         {load target buffer}
        push    di              {save original starting offset (in case != 0)}
        lds     si,inb          {load source buffer}
        add     si,4            {skip magic number}
        cld                     {make strings copy forward}
        mov     bx,OFFSET @SHR4table {prepare BX for XLAT later on}
        lodsw                   {load chunk size low 16-bit word}
        mov     bp,ax           {BP = size of compressed chunk}
        lodsw                   {load chunk size high 16-bit word}
        add     bp,si           {BP = threshold to stop decompression}
        or      ax,ax           {is high word non-zero?}
        jnz     @done           {If so, chunk too big or malformed, abort}

@starttoken:
        lodsb                   {grab token to AL}
        mov     dx,ax           {preserve packed token in DX}
  segcs xlat                    {unpack upper 4 bits, faster than SHR reg,cl}
        mov     cx,ax           {CX = unpacked literal length token}
        jcxz    @copymatches    {if CX = 0, no literals; try matches}
        cmp     al,$0F          {is it 15?}
        jne     @doliteralcopy1 {if so, build full length, else start copying}
@build1stcount:                 {this first count build is not the same}
        lodsb                   {fall-through jump as the one in the main loop}
        add     cx,ax           {because it is more likely that the very first}
        cmp     al,$FF          {length is 15 or more}
        je      @build1stcount
@doliteralcopy1:
        rep     movsb           {src and dst might overlap so do this by bytes}

{At this point, we might be done; all LZ4 data ends with five literals and the}
{offset token is ignored.  If we're at the end of our compressed chunk, stop.}

        cmp     si,bp           {are we at the end of our compressed chunk?}
        jae     @done           {if so, jump to exit; otherwise, process match}

@copymatches:
        lodsw                   {AX = match offset}
        xchg    dx,ax           {AX = packed token, DX = match offset}
        and     al,$0F          {unpack match length token}
        cmp     al,$0F          {is it 15?}
        xchg    cx,ax           {(doesn't affect flags); don't need ax any more}
        je      @buildmcount    {if not, start copying, otherwise build count}

@domatchcopy:
        cmp     dx,2            {if match offset=1 or 2, we're repeating a value}
        jbe     @domatchfill    {if so, perform RLE expansion optimally}
        push    ds
        xchg    si,ax           {ds:si saved}
        mov     si,di
        sub     si,dx
        mov     dx,es
        mov     ds,dx           {ds:si points at match; es:di points at dest}
        movsw                   
        movsw                   {minimum match is 4 bytes; move them ourselves}
        shr     cx,1
        rep     movsw           {cx contains count-4 so copy the rest}
        adc     cx,cx
        rep     movsb
        xchg    si,ax
        pop     ds              {ds:si restored}

@parsetoken:                    {CX always 0 here because of REP}
        xchg    cx,ax           {zero ah here to benefit other reg loads}
        lodsb                   {grab token to AL}
        mov     dx,ax           {preserve packed token in DX}
@copyliterals:                  {next 5 lines are 8088-optimal, do not rearrange}
  segcs xlat                    {unpack upper 4 bits, faster than SHR reg,cl}
        mov     cx,ax           {CX = unpacked literal length token}
        jcxz    @copymatches    {if CX = 0, no literals; try matches}
        cmp     al,$0F          {is it 15?}
        je      @buildlcount    {if so, build full length, else start copying}
@doliteralcopy:                 {src and dst might overlap so do this by bytes}
        rep     movsb           {if cx=0 nothing happens}

{At this point, we might be done; all LZ4 data ends with five literals and the}
{offset token is ignored.  If we're at the end of our compressed chunk, stop.}

@testformore:
        cmp     si,bp           {are we at the end of our compressed chunk?}
        jb      @copymatches    {if not, keep going}
        jmp     @done           {if so, end}

@domatchfill:
        je      @domatchfill2   {if DX=2, RLE by word, else by byte}
@domatchfill1:
        mov     al,es:[di-1]    {load byte we are filling with}
        mov     ah,al           {copy to ah so we can do 16-bit fills}
        stosw                   {minimum match is 4 bytes, so we fill four}
        stosw
        inc     cx              {round up for the shift}
        shr     cx,1            {CX = remaining (count+1)/2}
        rep     stosw           {includes odd byte - ok because LZ4 never ends with matches}
        adc     di,-1           {Adjust dest unless original count was even}
        jmp     @parsetoken     {continue decompressing}

@domatchfill2:
        mov     ax,es:[di-2]    {load word we are filling with}
        stosw                   {minimum match is 4 bytes, so we fill four}
        stosw
        inc     cx              {round up for the shift}
        shr     cx,1            {CX = remaining (count+1)/2}
        rep     stosw           {includes odd byte - ok because LZ4 never ends with matches}
        adc     di,-1           {Adjust dest unless original count was even}
        jmp     @parsetoken     {continue decompressing}

@buildlcount:                   {build full literal length count}
        lodsb                   {get next literal count byte}
        add     cx,ax           {increase count}
        cmp     al,$FF          {more count bytes to read?}
        je      @buildlcount
        jmp     @doliteralcopy

@buildmcount:                   {build full match length count - AX is 0}
        lodsb                   {get next literal count byte}
        add     cx,ax           {increase count}
        cmp     al,$FF          {more count bytes to read?}
        je      @buildmcount
        jmp     @domatchcopy

@done:
        pop     ax              {retrieve previous starting offset}
        sub     di,ax           {subtract prev offset from where we are now}
        xchg    ax,di           {AX = decompressed size}
        pop     bp              {restore compiler assumptions}
        pop     ds              {restore compiler assumptions}
end;

procedure nop1; assembler;
asm
  nop
end;

function lz4_decompress_small(inb,outb:pointer):word; assembler;
{
Same as LZ4_Decompress but optimized for size, not speed.  Still pretty fast,
although roughly 30% slower than lz4_decompress and RLE sequences are not
optimally handled.  Same Input, Output, and Trashes as lz4_decompress.
Assembles to 78 bytes. Thanks to Peter Ferrie for suggestions!
}
asm
        push    ds              {preserve compiler assumptions}
        les     di,outb         {load target buffer}
        push    di              {save original starting offset (in case != 0)}
        lds     si,inb          {load source buffer}
        cld                     {make strings copy forward}
        lodsw
        lodsw                   {skip magic number, smaller than "add si,4"}
        lodsw                   {load chunk size low 16-bit word}
        xchg    bx,ax           {BX = size of compressed chunk}
        add     bx,si           {BX = threshold to stop decompression}
        lodsw                   {load chunk size high 16-bit word}
        xchg    cx,ax           {set CX=0 so that AX=0 later}
        inc     cx              {is high word non-zero?}
        loop    @done           {If so, chunk too big or malformed, abort}

@parsetoken:                    {CX=0 here because of REP at end of loop}
        xchg    cx,ax           {zero ah here to benefit other reg loads}
        lodsb                   {grab token to AL}
        mov     dx,ax           {preserve packed token in DX}
@copyliterals:
        mov     cl,4
        shr     al,cl           {unpack upper 4 bits}
        call    @buildfullcount {build full literal count if necessary}
@doliteralcopy:                 {src and dst might overlap so do this by bytes}
        rep     movsb           {if cx=0 nothing happens}

{At this point, we might be done; all LZ4 data ends with five literals and the}
{offset token is ignored.  If we're at the end of our compressed chunk, stop.}

        cmp     si,bx           {are we at the end of our compressed chunk?}
        jae     @done           {if so, jump to exit; otherwise, process match}

@copymatches:
        lodsw                   {AX = match offset}
        xchg    dx,ax           {AX = packed token, DX = match offset}
        and     al,$0f          {unpack match length token}
        call    @buildfullcount {build full match count if necessary}
@domatchcopy:
        push    ds
        xchg    si,ax           {ds:si saved}
        mov     si,di
        sub     si,dx
        push    es
        pop     ds              {ds:si points at match; es:di points at dest}
	movsb
	movsb
	movsw                   {minmatch = 4}
        rep     movsb           {copy match run if any left}
        xchg    si,ax
        pop     ds              {ds:si restored}
        jmp     @parsetoken

@buildfullcount:
        mov     cx,ax           {CX = unpacked literal length token}
        cmp     al,$0F          {is it 15?}
        jne     @builddone      {if not, we have nothing to build}
@buildloop:
        lodsb                   {load a byte}
        add     cx,ax           {add it to the full count}
        cmp     al,$FF          {was it $FF?}
        je      @buildloop      {if so, keep going}
@builddone:
        ret

@done:
        pop     ax              {retrieve previous starting offset}
        sub     di,ax           {subtract prev offset from where we are now}
        xchg    ax,di           {AX = decompressed size}
        pop     ds              {restore compiler assumptions}
end;

procedure nop2; assembler;
asm
  nop
end;


{
The remainder of this file is a test harness for the above routines.
}

Function BuffersEqual(p1,p2:pointer;size:word):word; assembler;
{compares two buffers; reports 0 if equal, or offset where not}
asm
  push ds
  lds si,p1
  les di,p2
  mov cx,size
  xor ax,ax {assume test will be successful}
  cld
  repe cmpsb
  cmp cx,0
  je @done
  mov ax,size {if cx<>0 then buffers are not equal}
  sub ax,cx
  dec ax {to report offset instead of absolute location}
@done:
  pop ds
end;


const
  bufsize=65536-16; {turbo pascal heap limitation prevents full 64KiB}

type
  barray=array[0..bufsize-1] of byte;

var
  inf,outf:file;
  inbuf,outbuf:^barray;
  ins,outs:string;
  oldsize,newsize,decompsize:word;
  lzdecomptime,copytime:longint;

begin
  nop1; nop2;
  if paramcount<1 then begin
    writeln('Usage: lz4decom filename    (do NOT provide .lz4 extension)');
    halt(1);
  end;
  getmem(inbuf,bufsize); fillchar(inbuf^,sizeof(inbuf),$ff);
  getmem(outbuf,bufsize); fillchar(outbuf^,sizeof(outbuf),$ff);

  ins:=paramstr(1)+'.lz4';
  outs:=paramstr(1)+'.out';
  assign(inf,ins);
  reset(inf,1);
  blockread(inf,inbuf^,bufsize,oldsize);
  close(inf);

  write('Decompressing ',ins,'... ');
  _lztimeron;
  decompsize:=lz4_decompress(inbuf,outbuf);
  _lztimeroff;
  writeln('Done.  ',decompsize,' bytes of output in: ',_lztimercount);

  writeln('(small)');
  write('Decompressing ',ins,'... ');
  _lztimeron;
  decompsize:=lz4_decompress_small(inbuf,outbuf);
  _lztimeroff;
  writeln('Done.  ',decompsize,' bytes of output in: ',_lztimercount);

  lzdecomptime:=_lztimercount;
  if decompsize>0
    then begin
      {writeln('Writing data to 'paramstr(1),'.out');}
      assign(outf,outs);
      rewrite(outf,1);
      blockwrite(outf,outbuf^,decompsize);
      close(outf);
    end else writeln('Decompression failed!');

  {compare to REP MOVS by copying output buffer to input buffer}
  _lztimeron;
  asm
    mov cx,decompsize
    les di,outbuf
    push ds
    lds si,inbuf
    shr cx,1
    rep movsw
    adc cx,cx
    rep movsb
    pop ds
  end;
  _lztimeroff;
  copytime:=_lztimercount;
  writeln('memcpy took: ',copytime,'. decomp is ',
  lzdecomptime/copytime:3:2
  ,'x slower than memcopy.');

end.
